(load "mscvns.scm")

;;; basic tests
(eval-prog '10)
(eval (parse 'x) (new-env (list 'x) (list 15)))
(eval (parse '(+ 1 x)) (new-env (list 'x) (list 15)))
(eval-prog '(fun (x y) (+ x y)))
(eval (parse '(let ((y 10)) (+ y x))) (new-env (list 'x) (list 15)))
(eval-prog '(let ((f (fun (x y) (+ x y))))
          (f 3 4)))

;; side effects
(eval (parse '(let ((y 10)) 
                (+ y x)
                (set! x 2)
                (+ y x))) (new-env (list 'x) (list 15)))

;; call-by-value semantics
(eval-prog '(let ((v 0))
                (let ((f (fun (y) (set! y 5))))
                  (f v)
                  v)))


(printf "----call-by-value semantics for contextual values too----\n")
(eval-prog '(let ((v (make-cv (fun () 1) 0)))
                  (let ((f (fun (y) (set! y 5))))
                    (f v)
                    v)))

(printf "----interaction with function application / call-by-value ----\n")
(eval-prog '(let ((ctx 1))
                (let ((printit (fun (arg)
                                    (set! ctx 1)
                                    (printf "arg is ~s\n" arg)
                                    (set! arg 6)
                                    (printf "arg is ~s\n" arg)
                                    (set! ctx 2)
                                    (printf "arg is ~s\n" arg)
                                    ))) 
                  (let ((x (make-cv (fun () ctx) 10)))
                    (set! ctx 2)
                    (set! x 5)
                    (printit x)
                    (printf "x is ~s\n" x)
                    ))))

(printf "----call-by-value semantics for compound contextual values----\n")
(eval-prog '(let ((v (make-cv (fun () 1) 0)))
                  (set! v 1)
                  (let ((w (make-cv (fun () 2) v)))
                    (set! w 2)
                    (let ((f (fun (y) 
                                  (set! y 5) 
                                  (cv-trace y))))
                      (f w)
                      (cv-trace w)))))

(printf "----let and call-by-value----\n")
(eval-prog '(let ((ctx 1))
                (let ((x (make-cv (fun () ctx) 10)))
                  (set! ctx 2)
                  (set! x 5)
                  (let ((y x))
                    (set! y 3)
                    (printf "x is ~s\n" x)
                    (cv-trace x)
                    (cv-trace y)))))


;; playing with contextual values
(printf "--------\n")
(eval-prog '(let ((ctx 1))
                (let ((x (make-cv (fun () ctx) 10)))
                  (cv-trace x)
                  (set! ctx 2)
                  (set! x 3)
                  (cv-trace x)
                  (printf "x is ~a\n" x)
                  (set! ctx 1)
                  (cv-trace x)
                  (printf "x is ~a\n" x)
                  (set! x 5)
                  (cv-trace x)
                  (printf "x is ~a\n" x)
                  (set! ctx 2)
                  (printf "x is ~a\n" x)
                  )))

(printf "--------\n")
(eval-prog '(let ((ctx 1) (ctx2 'a))
                (let ((x 2))
                  (set! x (make-cv (fun () ctx) x))
                  (let ((y (make-cv (fun () ctx2) 0)))
                    (set! x 3)               ;;; 1a
                    (set! y x)
                    (printf "x is ~a\n" x)
                    (printf "y is ~a\n\n" y)
                    (set! ctx2 'b)          ;;; 1b
                    (printf "x is ~a\n" x)
                    (printf "y is ~a\n\n" y)                    
                    (set! ctx 2)          ;;; 2b
                    (printf "x is ~a\n" x)
                    (printf "y is ~a\n\n" y)
                    (set! ctx2 'a)          ;;; 2a
                    (printf "x is ~a\n" x)
                    (printf "y is ~a\n\n" y)
                  ))))



;; playing with scoped regions
(printf "--------\n")
(eval-prog '(let ((ctx 1))
                (let ((ctx-fun (fun () ctx)))
                  (with-scope ;--------------------
                   (let ((x 0))
                     (printf "x is ~a\n" x)
                     (set! x 1)
                     (printf "x is ~a\n" x)
                     (set! ctx 2)
                     (cv-trace x)
                     (scoped ctx-fun ;-------------
                             (printf "x is ~a\n" x)
                             (set! x 3)
                             (cv-trace x)
                             (printf "x is ~a\n" x)
                             (set! ctx 1)
                             (set! x 2)
                             (cv-trace x)
                             (printf "x is ~a\n" x)
                             )
                     (set! ctx 1)
                     (printf "x is ~a\n" x)
                     (set! ctx 2)
                     (printf "x is ~a\n" x)
                     (set! ctx 3)
                     (printf "x is ~a\n" x)
                     )))))

(printf "--------\n")
(eval-prog 
       '(let ((ctx1 1) (ctx2 'a))
          (let ((ctx1-fun (fun () ctx1))
                (ctx2-fun (fun () ctx2)))
            (with-scope ;--------------------
             (let ((x 0))
               (scoped ctx1-fun
                       (set! x 5)
                       (cv-trace x))
               (scoped ctx2-fun
                       (set! x 2)
                       (cv-trace x))
               (cv-trace x)
               (printf "x is ~a\n" x)
               (set! ctx1 1)
               (set! ctx2 'a)
               (printf "x is ~a\n" x)
               (set! ctx1 1)
               (set! ctx2 'b)
               (printf "x is ~a\n" x)
               (set! ctx1 2)
               (set! ctx2 'a)
               (printf "x is ~a\n" x)
               (set! ctx1 2)
               (set! ctx2 'b)
               (printf "x is ~a\n" x))))))

(printf "------h--\n")
(eval-prog 
       '(let ((ctx1 1) (ctx2 'a))
          (let ((ctx1-fun (fun () ctx1))
                (ctx2-fun (fun () ctx2)))
            (with-scope ;--------------------
             (let ((x 0))
               (scoped ctx1-fun
                       (set! x 5)
                       (scoped ctx1-fun 
                               (set! ctx1 2)
                               (set! x 2))
                       (cv-trace x))
               (set! ctx1 1)
               (printf "x is ~a\n" x)
               (set! ctx1 2)
               (printf "x is ~a\n" x)
               )))))

(printf "--------\n")
(eval-prog 
       '(let ((ctx1 1) (ctx2 'a))
          (let ((ctx1-fun (fun () ctx1))
                (ctx2-fun (fun () ctx2)))
            (with-scope ;--------------------
             (let ((x 0))
               (scoped ctx1-fun
                       (set! x 5)
                       (printf "x is ~a\n" x)
                       (cv-trace x))
               (scoped ctx1-fun
                       (set! ctx1 2)
                       (set! x 4)
                       (printf "x is ~a\n" x)
                       (cv-trace x))
               (set! ctx1 1)
               (printf "x is ~a\n" x)
               (set! ctx1 2)
               (printf "x is ~a\n" x)
               )))))

(printf "--------\n")
(eval-prog 
       '(let ((ctx1 1) (ctx2 'a))
          (let ((ctx1-fun (fun () ctx1))
                (ctx2-fun (fun () ctx2)))
            (with-scope ;--------------------
             (let ((x 0))
               (scoped ctx1-fun
                       (set! x 5)
                       (cv-trace x)
                       (scoped ctx2-fun 
                               (set! ctx1 2)
                               (set! x 2)
                               (cv-trace x)))
               (cv-trace x)
               (set! ctx2 'a)
               (printf "x is ~a\n" x)
               (set! ctx1 1)
               (printf "x is ~a\n" x)
               )))))

(printf "--------\n")
(eval-prog 
       '(let ((ctx1 1) (ctx2 'a))
          (let ((ctx1-fun (fun () ctx1))
                (ctx2-fun (fun () ctx2)))
            (with-scope ;--------------------
             (let ((x 0))
               (scoped ctx1-fun
                       (set! x 5)
                       (cv-trace x))
               (scoped ctx2-fun 
                       (set! x 2))
             (cv-trace x)
             (set! ctx2 'b)
             (printf "x is ~a\n" x)
             (set! ctx1 1)
             (printf "x is ~a\n" x)
             )))))

(eval-prog
       '(let ((context 0))
          (let ((ctx-fun (fun () context)))
            (with-scope 
             (let ((x 0))
               (set! x 2)
               (cv-trace x)
               (scoped ctx-fun
                       (set! ctx-fun (fun () 1))
                       (set! x 1)
                       (set! context 2)
                       (set! x 3)
                       (cv-trace x))
               (set! ctx-fun (fun () context))
               (set! context 1)
               x
               )))))

(printf "---------\n")
(eval-prog 
       '(let ((user "totoro") (thread-id 1))
          (let ((userctx (fun () user))
                (threadctx (fun () thread-id)))
            (with-scope
             (let ((counter 0))
               (scoped threadctx
                       (scoped userctx
                               (set! counter 100)))
               (cv-trace counter))))))

(printf "---------\n")
(eval-prog 
       '(let ((user "totoro") (thread-id 1))
          (let ((userctx (fun () user))
                (threadctx (fun () thread-id)))
            (with-scope
             (let ((counter 0))
               (scoped threadctx
                       (scoped userctx
                               (set! counter 100)))
               (set! thread-id 2)
               (set! user "bilbo")
               (scoped threadctx
                       (set! counter 50)) 
               (cv-trace counter))))))
               